#
#
# AUTOGENERATED, DO NOT EDIT
#
# From version: 4.11
#

i386 = {
    0: "sys_restart_syscall",
    1: "sys_exit",
    2: "sys_fork",
    3: "sys_read",
    4: "sys_write",
    5: "sys_open",
    6: "sys_close",
    7: "sys_waitpid",
    8: "sys_creat",
    9: "sys_link",
    10: "sys_unlink",
    11: "sys_execve",
    12: "sys_chdir",
    13: "sys_time",
    14: "sys_mknod",
    15: "sys_chmod",
    16: "sys_lchown16",
    18: "sys_stat",
    19: "sys_lseek",
    20: "sys_getpid",
    21: "sys_mount",
    22: "sys_oldumount",
    23: "sys_setuid16",
    24: "sys_getuid16",
    25: "sys_stime",
    26: "sys_ptrace",
    27: "sys_alarm",
    28: "sys_fstat",
    29: "sys_pause",
    30: "sys_utime",
    33: "sys_access",
    34: "sys_nice",
    36: "sys_sync",
    37: "sys_kill",
    38: "sys_rename",
    39: "sys_mkdir",
    40: "sys_rmdir",
    41: "sys_dup",
    42: "sys_pipe",
    43: "sys_times",
    45: "sys_brk",
    46: "sys_setgid16",
    47: "sys_getgid16",
    48: "sys_signal",
    49: "sys_geteuid16",
    50: "sys_getegid16",
    51: "sys_acct",
    52: "sys_umount",
    54: "sys_ioctl",
    55: "sys_fcntl",
    57: "sys_setpgid",
    59: "sys_olduname",
    60: "sys_umask",
    61: "sys_chroot",
    62: "sys_ustat",
    63: "sys_dup2",
    64: "sys_getppid",
    65: "sys_getpgrp",
    66: "sys_setsid",
    67: "sys_sigaction",
    68: "sys_sgetmask",
    69: "sys_ssetmask",
    70: "sys_setreuid16",
    71: "sys_setregid16",
    72: "sys_sigsuspend",
    73: "sys_sigpending",
    74: "sys_sethostname",
    75: "sys_setrlimit",
    76: "sys_old_getrlimit",
    77: "sys_getrusage",
    78: "sys_gettimeofday",
    79: "sys_settimeofday",
    80: "sys_getgroups16",
    81: "sys_setgroups16",
    82: "sys_old_select",
    83: "sys_symlink",
    84: "sys_lstat",
    85: "sys_readlink",
    86: "sys_uselib",
    87: "sys_swapon",
    88: "sys_reboot",
    89: "sys_old_readdir",
    90: "sys_old_mmap",
    91: "sys_munmap",
    92: "sys_truncate",
    93: "sys_ftruncate",
    94: "sys_fchmod",
    95: "sys_fchown16",
    96: "sys_getpriority",
    97: "sys_setpriority",
    99: "sys_statfs",
    100: "sys_fstatfs",
    101: "sys_ioperm",
    102: "sys_socketcall",
    103: "sys_syslog",
    104: "sys_setitimer",
    105: "sys_getitimer",
    106: "sys_newstat",
    107: "sys_newlstat",
    108: "sys_newfstat",
    109: "sys_uname",
    110: "sys_iopl",
    111: "sys_vhangup",
    113: "sys_vm86old",
    114: "sys_wait4",
    115: "sys_swapoff",
    116: "sys_sysinfo",
    117: "sys_ipc",
    118: "sys_fsync",
    119: "sys_sigreturn",
    120: "sys_clone",
    121: "sys_setdomainname",
    122: "sys_newuname",
    123: "sys_modify_ldt",
    124: "sys_adjtimex",
    125: "sys_mprotect",
    126: "sys_sigprocmask",
    128: "sys_init_module",
    129: "sys_delete_module",
    131: "sys_quotactl",
    132: "sys_getpgid",
    133: "sys_fchdir",
    134: "sys_bdflush",
    135: "sys_sysfs",
    136: "sys_personality",
    138: "sys_setfsuid16",
    139: "sys_setfsgid16",
    140: "sys_llseek",
    141: "sys_getdents",
    142: "sys_select",
    143: "sys_flock",
    144: "sys_msync",
    145: "sys_readv",
    146: "sys_writev",
    147: "sys_getsid",
    148: "sys_fdatasync",
    149: "sys_sysctl",
    150: "sys_mlock",
    151: "sys_munlock",
    152: "sys_mlockall",
    153: "sys_munlockall",
    154: "sys_sched_setparam",
    155: "sys_sched_getparam",
    156: "sys_sched_setscheduler",
    157: "sys_sched_getscheduler",
    158: "sys_sched_yield",
    159: "sys_sched_get_priority_max",
    160: "sys_sched_get_priority_min",
    161: "sys_sched_rr_get_interval",
    162: "sys_nanosleep",
    163: "sys_mremap",
    164: "sys_setresuid16",
    165: "sys_getresuid16",
    166: "sys_vm86",
    168: "sys_poll",
    170: "sys_setresgid16",
    171: "sys_getresgid16",
    172: "sys_prctl",
    173: "sys_rt_sigreturn",
    174: "sys_rt_sigaction",
    175: "sys_rt_sigprocmask",
    176: "sys_rt_sigpending",
    177: "sys_rt_sigtimedwait",
    178: "sys_rt_sigqueueinfo",
    179: "sys_rt_sigsuspend",
    180: "sys_pread64",
    181: "sys_pwrite64",
    182: "sys_chown16",
    183: "sys_getcwd",
    184: "sys_capget",
    185: "sys_capset",
    186: "sys_sigaltstack",
    187: "sys_sendfile",
    190: "sys_vfork",
    191: "sys_getrlimit",
    192: "sys_mmap_pgoff",
    193: "sys_truncate64",
    194: "sys_ftruncate64",
    195: "sys_stat64",
    196: "sys_lstat64",
    197: "sys_fstat64",
    198: "sys_lchown",
    199: "sys_getuid",
    200: "sys_getgid",
    201: "sys_geteuid",
    202: "sys_getegid",
    203: "sys_setreuid",
    204: "sys_setregid",
    205: "sys_getgroups",
    206: "sys_setgroups",
    207: "sys_fchown",
    208: "sys_setresuid",
    209: "sys_getresuid",
    210: "sys_setresgid",
    211: "sys_getresgid",
    212: "sys_chown",
    213: "sys_setuid",
    214: "sys_setgid",
    215: "sys_setfsuid",
    216: "sys_setfsgid",
    217: "sys_pivot_root",
    218: "sys_mincore",
    219: "sys_madvise",
    220: "sys_getdents64",
    221: "sys_fcntl64",
    224: "sys_gettid",
    225: "sys_readahead",
    226: "sys_setxattr",
    227: "sys_lsetxattr",
    228: "sys_fsetxattr",
    229: "sys_getxattr",
    230: "sys_lgetxattr",
    231: "sys_fgetxattr",
    232: "sys_listxattr",
    233: "sys_llistxattr",
    234: "sys_flistxattr",
    235: "sys_removexattr",
    236: "sys_lremovexattr",
    237: "sys_fremovexattr",
    238: "sys_tkill",
    239: "sys_sendfile64",
    240: "sys_futex",
    241: "sys_sched_setaffinity",
    242: "sys_sched_getaffinity",
    243: "sys_set_thread_area",
    244: "sys_get_thread_area",
    245: "sys_io_setup",
    246: "sys_io_destroy",
    247: "sys_io_getevents",
    248: "sys_io_submit",
    249: "sys_io_cancel",
    250: "sys_fadvise64",
    252: "sys_exit_group",
    253: "sys_lookup_dcookie",
    254: "sys_epoll_create",
    255: "sys_epoll_ctl",
    256: "sys_epoll_wait",
    257: "sys_remap_file_pages",
    258: "sys_set_tid_address",
    259: "sys_timer_create",
    260: "sys_timer_settime",
    261: "sys_timer_gettime",
    262: "sys_timer_getoverrun",
    263: "sys_timer_delete",
    264: "sys_clock_settime",
    265: "sys_clock_gettime",
    266: "sys_clock_getres",
    267: "sys_clock_nanosleep",
    268: "sys_statfs64",
    269: "sys_fstatfs64",
    270: "sys_tgkill",
    271: "sys_utimes",
    272: "sys_fadvise64_64",
    274: "sys_mbind",
    275: "sys_get_mempolicy",
    276: "sys_set_mempolicy",
    277: "sys_mq_open",
    278: "sys_mq_unlink",
    279: "sys_mq_timedsend",
    280: "sys_mq_timedreceive",
    281: "sys_mq_notify",
    282: "sys_mq_getsetattr",
    283: "sys_kexec_load",
    284: "sys_waitid",
    286: "sys_add_key",
    287: "sys_request_key",
    288: "sys_keyctl",
    289: "sys_ioprio_set",
    290: "sys_ioprio_get",
    291: "sys_inotify_init",
    292: "sys_inotify_add_watch",
    293: "sys_inotify_rm_watch",
    294: "sys_migrate_pages",
    295: "sys_openat",
    296: "sys_mkdirat",
    297: "sys_mknodat",
    298: "sys_fchownat",
    299: "sys_futimesat",
    300: "sys_fstatat64",
    301: "sys_unlinkat",
    302: "sys_renameat",
    303: "sys_linkat",
    304: "sys_symlinkat",
    305: "sys_readlinkat",
    306: "sys_fchmodat",
    307: "sys_faccessat",
    308: "sys_pselect6",
    309: "sys_ppoll",
    310: "sys_unshare",
    311: "sys_set_robust_list",
    312: "sys_get_robust_list",
    313: "sys_splice",
    314: "sys_sync_file_range",
    315: "sys_tee",
    316: "sys_vmsplice",
    317: "sys_move_pages",
    318: "sys_getcpu",
    319: "sys_epoll_pwait",
    320: "sys_utimensat",
    321: "sys_signalfd",
    322: "sys_timerfd_create",
    323: "sys_eventfd",
    324: "sys_fallocate",
    325: "sys_timerfd_settime",
    326: "sys_timerfd_gettime",
    327: "sys_signalfd4",
    328: "sys_eventfd2",
    329: "sys_epoll_create1",
    330: "sys_dup3",
    331: "sys_pipe2",
    332: "sys_inotify_init1",
    333: "sys_preadv",
    334: "sys_pwritev",
    335: "sys_rt_tgsigqueueinfo",
    336: "sys_perf_event_open",
    337: "sys_recvmmsg",
    338: "sys_fanotify_init",
    339: "sys_fanotify_mark",
    340: "sys_prlimit64",
    341: "sys_name_to_handle_at",
    342: "sys_open_by_handle_at",
    343: "sys_clock_adjtime",
    344: "sys_syncfs",
    345: "sys_sendmmsg",
    346: "sys_setns",
    347: "sys_process_vm_readv",
    348: "sys_process_vm_writev",
    349: "sys_kcmp",
    350: "sys_finit_module",
    351: "sys_sched_setattr",
    352: "sys_sched_getattr",
    353: "sys_renameat2",
    354: "sys_seccomp",
    355: "sys_getrandom",
    356: "sys_memfd_create",
    357: "sys_bpf",
    358: "sys_execveat",
    359: "sys_socket",
    360: "sys_socketpair",
    361: "sys_bind",
    362: "sys_connect",
    363: "sys_listen",
    364: "sys_accept4",
    365: "sys_getsockopt",
    366: "sys_setsockopt",
    367: "sys_getsockname",
    368: "sys_getpeername",
    369: "sys_sendto",
    370: "sys_sendmsg",
    371: "sys_recvfrom",
    372: "sys_recvmsg",
    373: "sys_shutdown",
    374: "sys_userfaultfd",
    375: "sys_membarrier",
    376: "sys_mlock2",
    377: "sys_copy_file_range",
    378: "sys_preadv2",
    379: "sys_pwritev2",
    380: "sys_pkey_mprotect",
    381: "sys_pkey_alloc",
    382: "sys_pkey_free",
    383: "sys_statx",
}
amd64 = {
    0: "sys_read",
    1: "sys_write",
    2: "sys_open",
    3: "sys_close",
    4: "sys_newstat",
    5: "sys_newfstat",
    6: "sys_newlstat",
    7: "sys_poll",
    8: "sys_lseek",
    9: "sys_mmap",
    10: "sys_mprotect",
    11: "sys_munmap",
    12: "sys_brk",
    13: "sys_rt_sigaction",
    14: "sys_rt_sigprocmask",
    15: "sys_rt_sigreturn/ptregs",
    16: "sys_ioctl",
    17: "sys_pread64",
    18: "sys_pwrite64",
    19: "sys_readv",
    20: "sys_writev",
    21: "sys_access",
    22: "sys_pipe",
    23: "sys_select",
    24: "sys_sched_yield",
    25: "sys_mremap",
    26: "sys_msync",
    27: "sys_mincore",
    28: "sys_madvise",
    29: "sys_shmget",
    30: "sys_shmat",
    31: "sys_shmctl",
    32: "sys_dup",
    33: "sys_dup2",
    34: "sys_pause",
    35: "sys_nanosleep",
    36: "sys_getitimer",
    37: "sys_alarm",
    38: "sys_setitimer",
    39: "sys_getpid",
    40: "sys_sendfile64",
    41: "sys_socket",
    42: "sys_connect",
    43: "sys_accept",
    44: "sys_sendto",
    45: "sys_recvfrom",
    46: "sys_sendmsg",
    47: "sys_recvmsg",
    48: "sys_shutdown",
    49: "sys_bind",
    50: "sys_listen",
    51: "sys_getsockname",
    52: "sys_getpeername",
    53: "sys_socketpair",
    54: "sys_setsockopt",
    55: "sys_getsockopt",
    56: "sys_clone_ptregs",
    57: "sys_fork/ptregs",
    58: "sys_vfork/ptregs",
    59: "sys_execve/ptregs",
    60: "sys_exit",
    61: "sys_wait4",
    62: "sys_kill",
    63: "sys_newuname",
    64: "sys_semget",
    65: "sys_semop",
    66: "sys_semctl",
    67: "sys_shmdt",
    68: "sys_msgget",
    69: "sys_msgsnd",
    70: "sys_msgrcv",
    71: "sys_msgctl",
    72: "sys_fcntl",
    73: "sys_flock",
    74: "sys_fsync",
    75: "sys_fdatasync",
    76: "sys_truncate",
    77: "sys_ftruncate",
    78: "sys_getdents",
    79: "sys_getcwd",
    80: "sys_chdir",
    81: "sys_fchdir",
    82: "sys_rename",
    83: "sys_mkdir",
    84: "sys_rmdir",
    85: "sys_creat",
    86: "sys_link",
    87: "sys_unlink",
    88: "sys_symlink",
    89: "sys_readlink",
    90: "sys_chmod",
    91: "sys_fchmod",
    92: "sys_chown",
    93: "sys_fchown",
    94: "sys_lchown",
    95: "sys_umask",
    96: "sys_gettimeofday",
    97: "sys_getrlimit",
    98: "sys_getrusage",
    99: "sys_sysinfo",
    100: "sys_times",
    101: "sys_ptrace",
    102: "sys_getuid",
    103: "sys_syslog",
    104: "sys_getgid",
    105: "sys_setuid",
    106: "sys_setgid",
    107: "sys_geteuid",
    108: "sys_getegid",
    109: "sys_setpgid",
    110: "sys_getppid",
    111: "sys_getpgrp",
    112: "sys_setsid",
    113: "sys_setreuid",
    114: "sys_setregid",
    115: "sys_getgroups",
    116: "sys_setgroups",
    117: "sys_setresuid",
    118: "sys_getresuid",
    119: "sys_setresgid",
    120: "sys_getresgid",
    121: "sys_getpgid",
    122: "sys_setfsuid",
    123: "sys_setfsgid",
    124: "sys_getsid",
    125: "sys_capget",
    126: "sys_capset",
    127: "sys_rt_sigpending",
    128: "sys_rt_sigtimedwait",
    129: "sys_rt_sigqueueinfo",
    130: "sys_rt_sigsuspend",
    131: "sys_sigaltstack",
    132: "sys_utime",
    133: "sys_mknod",
    135: "sys_personality",
    136: "sys_ustat",
    137: "sys_statfs",
    138: "sys_fstatfs",
    139: "sys_sysfs",
    140: "sys_getpriority",
    141: "sys_setpriority",
    142: "sys_sched_setparam",
    143: "sys_sched_getparam",
    144: "sys_sched_setscheduler",
    145: "sys_sched_getscheduler",
    146: "sys_sched_get_priority_max",
    147: "sys_sched_get_priority_min",
    148: "sys_sched_rr_get_interval",
    149: "sys_mlock",
    150: "sys_munlock",
    151: "sys_mlockall",
    152: "sys_munlockall",
    153: "sys_vhangup",
    154: "sys_modify_ldt",
    155: "sys_pivot_root",
    156: "sys_sysctl",
    157: "sys_prctl",
    158: "sys_arch_prctl",
    159: "sys_adjtimex",
    160: "sys_setrlimit",
    161: "sys_chroot",
    162: "sys_sync",
    163: "sys_acct",
    164: "sys_settimeofday",
    165: "sys_mount",
    166: "sys_umount",
    167: "sys_swapon",
    168: "sys_swapoff",
    169: "sys_reboot",
    170: "sys_sethostname",
    171: "sys_setdomainname",
    172: "sys_iopl/ptregs",
    173: "sys_ioperm",
    175: "sys_init_module",
    176: "sys_delete_module",
    179: "sys_quotactl",
    186: "sys_gettid",
    187: "sys_readahead",
    188: "sys_setxattr",
    189: "sys_lsetxattr",
    190: "sys_fsetxattr",
    191: "sys_getxattr",
    192: "sys_lgetxattr",
    193: "sys_fgetxattr",
    194: "sys_listxattr",
    195: "sys_llistxattr",
    196: "sys_flistxattr",
    197: "sys_removexattr",
    198: "sys_lremovexattr",
    199: "sys_fremovexattr",
    200: "sys_tkill",
    201: "sys_time",
    202: "sys_futex",
    203: "sys_sched_setaffinity",
    204: "sys_sched_getaffinity",
    206: "sys_io_setup",
    207: "sys_io_destroy",
    208: "sys_io_getevents",
    209: "sys_io_submit",
    210: "sys_io_cancel",
    212: "sys_lookup_dcookie",
    213: "sys_epoll_create",
    216: "sys_remap_file_pages",
    217: "sys_getdents64",
    218: "sys_set_tid_address",
    219: "sys_restart_syscall",
    220: "sys_semtimedop",
    221: "sys_fadvise64",
    222: "sys_timer_create",
    223: "sys_timer_settime",
    224: "sys_timer_gettime",
    225: "sys_timer_getoverrun",
    226: "sys_timer_delete",
    227: "sys_clock_settime",
    228: "sys_clock_gettime",
    229: "sys_clock_getres",
    230: "sys_clock_nanosleep",
    231: "sys_exit_group",
    232: "sys_epoll_wait",
    233: "sys_epoll_ctl",
    234: "sys_tgkill",
    235: "sys_utimes",
    237: "sys_mbind",
    238: "sys_set_mempolicy",
    239: "sys_get_mempolicy",
    240: "sys_mq_open",
    241: "sys_mq_unlink",
    242: "sys_mq_timedsend",
    243: "sys_mq_timedreceive",
    244: "sys_mq_notify",
    245: "sys_mq_getsetattr",
    246: "sys_kexec_load",
    247: "sys_waitid",
    248: "sys_add_key",
    249: "sys_request_key",
    250: "sys_keyctl",
    251: "sys_ioprio_set",
    252: "sys_ioprio_get",
    253: "sys_inotify_init",
    254: "sys_inotify_add_watch",
    255: "sys_inotify_rm_watch",
    256: "sys_migrate_pages",
    257: "sys_openat",
    258: "sys_mkdirat",
    259: "sys_mknodat",
    260: "sys_fchownat",
    261: "sys_futimesat",
    262: "sys_newfstatat",
    263: "sys_unlinkat",
    264: "sys_renameat",
    265: "sys_linkat",
    266: "sys_symlinkat",
    267: "sys_readlinkat",
    268: "sys_fchmodat",
    269: "sys_faccessat",
    270: "sys_pselect6",
    271: "sys_ppoll",
    272: "sys_unshare",
    273: "sys_set_robust_list",
    274: "sys_get_robust_list",
    275: "sys_splice",
    276: "sys_tee",
    277: "sys_sync_file_range",
    278: "sys_vmsplice",
    279: "sys_move_pages",
    280: "sys_utimensat",
    281: "sys_epoll_pwait",
    282: "sys_signalfd",
    283: "sys_timerfd_create",
    284: "sys_eventfd",
    285: "sys_fallocate",
    286: "sys_timerfd_settime",
    287: "sys_timerfd_gettime",
    288: "sys_accept4",
    289: "sys_signalfd4",
    290: "sys_eventfd2",
    291: "sys_epoll_create1",
    292: "sys_dup3",
    293: "sys_pipe2",
    294: "sys_inotify_init1",
    295: "sys_preadv",
    296: "sys_pwritev",
    297: "sys_rt_tgsigqueueinfo",
    298: "sys_perf_event_open",
    299: "sys_recvmmsg",
    300: "sys_fanotify_init",
    301: "sys_fanotify_mark",
    302: "sys_prlimit64",
    303: "sys_name_to_handle_at",
    304: "sys_open_by_handle_at",
    305: "sys_clock_adjtime",
    306: "sys_syncfs",
    307: "sys_sendmmsg",
    308: "sys_setns",
    309: "sys_getcpu",
    310: "sys_process_vm_readv",
    311: "sys_process_vm_writev",
    312: "sys_kcmp",
    313: "sys_finit_module",
    314: "sys_sched_setattr",
    315: "sys_sched_getattr",
    316: "sys_renameat2",
    317: "sys_seccomp",
    318: "sys_getrandom",
    319: "sys_memfd_create",
    320: "sys_kexec_file_load",
    321: "sys_bpf",
    322: "sys_execveat/ptregs",
    323: "sys_userfaultfd",
    324: "sys_membarrier",
    325: "sys_mlock2",
    326: "sys_copy_file_range",
    327: "sys_preadv2",
    328: "sys_pwritev2",
    329: "sys_pkey_mprotect",
    330: "sys_pkey_alloc",
    331: "sys_pkey_free",
    332: "sys_statx",
    512: "compat_sys_rt_sigaction",
    513: "sys32_x32_rt_sigreturn",
    514: "compat_sys_ioctl",
    515: "compat_sys_readv",
    516: "compat_sys_writev",
    517: "compat_sys_recvfrom",
    518: "compat_sys_sendmsg",
    519: "compat_sys_recvmsg",
    520: "compat_sys_execve/ptregs",
    521: "compat_sys_ptrace",
    522: "compat_sys_rt_sigpending",
    523: "compat_sys_rt_sigtimedwait",
    524: "compat_sys_rt_sigqueueinfo",
    525: "compat_sys_sigaltstack",
    526: "compat_sys_timer_create",
    527: "compat_sys_mq_notify",
    528: "compat_sys_kexec_load",
    529: "compat_sys_waitid",
    530: "compat_sys_set_robust_list",
    531: "compat_sys_get_robust_list",
    532: "compat_sys_vmsplice",
    533: "compat_sys_move_pages",
    534: "compat_sys_preadv64",
    535: "compat_sys_pwritev64",
    536: "compat_sys_rt_tgsigqueueinfo",
    537: "compat_sys_recvmmsg",
    538: "compat_sys_sendmmsg",
    539: "compat_sys_process_vm_readv",
    540: "compat_sys_process_vm_writev",
    541: "compat_sys_setsockopt",
    542: "compat_sys_getsockopt",
    543: "compat_sys_io_setup",
    544: "compat_sys_io_submit",
    545: "compat_sys_execveat/ptregs",
    546: "compat_sys_preadv64v2",
    547: "compat_sys_pwritev64v2",
}
armv7 = {
    0: "sys_restart_syscall",
    1: "sys_exit",
    2: "sys_fork",
    3: "sys_read",
    4: "sys_write",
    5: "sys_open",
    6: "sys_close",
    8: "sys_creat",
    9: "sys_link",
    10: "sys_unlink",
    11: "sys_execve",
    12: "sys_chdir",
    13: "sys_time",
    14: "sys_mknod",
    15: "sys_chmod",
    16: "sys_lchown16",
    19: "sys_lseek",
    20: "sys_getpid",
    21: "sys_mount",
    22: "sys_oldumount",
    23: "sys_setuid16",
    24: "sys_getuid16",
    25: "sys_stime",
    26: "sys_ptrace",
    27: "sys_alarm",
    29: "sys_pause",
    30: "sys_utime",
    33: "sys_access",
    34: "sys_nice",
    36: "sys_sync",
    37: "sys_kill",
    38: "sys_rename",
    39: "sys_mkdir",
    40: "sys_rmdir",
    41: "sys_dup",
    42: "sys_pipe",
    43: "sys_times",
    45: "sys_brk",
    46: "sys_setgid16",
    47: "sys_getgid16",
    49: "sys_geteuid16",
    50: "sys_getegid16",
    51: "sys_acct",
    52: "sys_umount",
    54: "sys_ioctl",
    55: "sys_fcntl",
    57: "sys_setpgid",
    60: "sys_umask",
    61: "sys_chroot",
    62: "sys_ustat",
    63: "sys_dup2",
    64: "sys_getppid",
    65: "sys_getpgrp",
    66: "sys_setsid",
    67: "sys_sigaction",
    70: "sys_setreuid16",
    71: "sys_setregid16",
    72: "sys_sigsuspend",
    73: "sys_sigpending",
    74: "sys_sethostname",
    75: "sys_setrlimit",
    76: "sys_old_getrlimit",
    77: "sys_getrusage",
    78: "sys_gettimeofday",
    79: "sys_settimeofday",
    80: "sys_getgroups16",
    81: "sys_setgroups16",
    82: "sys_old_select",
    83: "sys_symlink",
    85: "sys_readlink",
    86: "sys_uselib",
    87: "sys_swapon",
    88: "sys_reboot",
    89: "sys_old_readdir",
    90: "sys_old_mmap",
    91: "sys_munmap",
    92: "sys_truncate",
    93: "sys_ftruncate",
    94: "sys_fchmod",
    95: "sys_fchown16",
    96: "sys_getpriority",
    97: "sys_setpriority",
    99: "sys_statfs",
    100: "sys_fstatfs",
    102: "sys_socketcall",
    103: "sys_syslog",
    104: "sys_setitimer",
    105: "sys_getitimer",
    106: "sys_newstat",
    107: "sys_newlstat",
    108: "sys_newfstat",
    111: "sys_vhangup",
    113: "sys_syscall",
    114: "sys_wait4",
    115: "sys_swapoff",
    116: "sys_sysinfo",
    117: "sys_ipc",
    118: "sys_fsync",
    119: "sys_sigreturn_wrapper",
    120: "sys_clone",
    121: "sys_setdomainname",
    122: "sys_newuname",
    124: "sys_adjtimex",
    125: "sys_mprotect",
    126: "sys_sigprocmask",
    128: "sys_init_module",
    129: "sys_delete_module",
    131: "sys_quotactl",
    132: "sys_getpgid",
    133: "sys_fchdir",
    134: "sys_bdflush",
    135: "sys_sysfs",
    136: "sys_personality",
    138: "sys_setfsuid16",
    139: "sys_setfsgid16",
    140: "sys_llseek",
    141: "sys_getdents",
    142: "sys_select",
    143: "sys_flock",
    144: "sys_msync",
    145: "sys_readv",
    146: "sys_writev",
    147: "sys_getsid",
    148: "sys_fdatasync",
    149: "sys_sysctl",
    150: "sys_mlock",
    151: "sys_munlock",
    152: "sys_mlockall",
    153: "sys_munlockall",
    154: "sys_sched_setparam",
    155: "sys_sched_getparam",
    156: "sys_sched_setscheduler",
    157: "sys_sched_getscheduler",
    158: "sys_sched_yield",
    159: "sys_sched_get_priority_max",
    160: "sys_sched_get_priority_min",
    161: "sys_sched_rr_get_interval",
    162: "sys_nanosleep",
    163: "sys_mremap",
    164: "sys_setresuid16",
    165: "sys_getresuid16",
    168: "sys_poll",
    170: "sys_setresgid16",
    171: "sys_getresgid16",
    172: "sys_prctl",
    173: "sys_rt_sigreturn_wrapper",
    174: "sys_rt_sigaction",
    175: "sys_rt_sigprocmask",
    176: "sys_rt_sigpending",
    177: "sys_rt_sigtimedwait",
    178: "sys_rt_sigqueueinfo",
    179: "sys_rt_sigsuspend",
    180: "sys_pread64",
    181: "sys_pwrite64",
    182: "sys_chown16",
    183: "sys_getcwd",
    184: "sys_capget",
    185: "sys_capset",
    186: "sys_sigaltstack",
    187: "sys_sendfile",
    190: "sys_vfork",
    191: "sys_getrlimit",
    192: "sys_mmap2",
    193: "sys_truncate64",
    194: "sys_ftruncate64",
    195: "sys_stat64",
    196: "sys_lstat64",
    197: "sys_fstat64",
    198: "sys_lchown",
    199: "sys_getuid",
    200: "sys_getgid",
    201: "sys_geteuid",
    202: "sys_getegid",
    203: "sys_setreuid",
    204: "sys_setregid",
    205: "sys_getgroups",
    206: "sys_setgroups",
    207: "sys_fchown",
    208: "sys_setresuid",
    209: "sys_getresuid",
    210: "sys_setresgid",
    211: "sys_getresgid",
    212: "sys_chown",
    213: "sys_setuid",
    214: "sys_setgid",
    215: "sys_setfsuid",
    216: "sys_setfsgid",
    217: "sys_getdents64",
    218: "sys_pivot_root",
    219: "sys_mincore",
    220: "sys_madvise",
    221: "sys_fcntl64",
    224: "sys_gettid",
    225: "sys_readahead",
    226: "sys_setxattr",
    227: "sys_lsetxattr",
    228: "sys_fsetxattr",
    229: "sys_getxattr",
    230: "sys_lgetxattr",
    231: "sys_fgetxattr",
    232: "sys_listxattr",
    233: "sys_llistxattr",
    234: "sys_flistxattr",
    235: "sys_removexattr",
    236: "sys_lremovexattr",
    237: "sys_fremovexattr",
    238: "sys_tkill",
    239: "sys_sendfile64",
    240: "sys_futex",
    241: "sys_sched_setaffinity",
    242: "sys_sched_getaffinity",
    243: "sys_io_setup",
    244: "sys_io_destroy",
    245: "sys_io_getevents",
    246: "sys_io_submit",
    247: "sys_io_cancel",
    248: "sys_exit_group",
    249: "sys_lookup_dcookie",
    250: "sys_epoll_create",
    251: "sys_epoll_ctl",
    252: "sys_epoll_wait",
    253: "sys_remap_file_pages",
    256: "sys_set_tid_address",
    257: "sys_timer_create",
    258: "sys_timer_settime",
    259: "sys_timer_gettime",
    260: "sys_timer_getoverrun",
    261: "sys_timer_delete",
    262: "sys_clock_settime",
    263: "sys_clock_gettime",
    264: "sys_clock_getres",
    265: "sys_clock_nanosleep",
    266: "sys_statfs64_wrapper",
    267: "sys_fstatfs64_wrapper",
    268: "sys_tgkill",
    269: "sys_utimes",
    270: "sys_arm_fadvise64_64",
    271: "sys_pciconfig_iobase",
    272: "sys_pciconfig_read",
    273: "sys_pciconfig_write",
    274: "sys_mq_open",
    275: "sys_mq_unlink",
    276: "sys_mq_timedsend",
    277: "sys_mq_timedreceive",
    278: "sys_mq_notify",
    279: "sys_mq_getsetattr",
    280: "sys_waitid",
    281: "sys_socket",
    282: "sys_bind",
    283: "sys_connect",
    284: "sys_listen",
    285: "sys_accept",
    286: "sys_getsockname",
    287: "sys_getpeername",
    288: "sys_socketpair",
    289: "sys_send",
    290: "sys_sendto",
    291: "sys_recv",
    292: "sys_recvfrom",
    293: "sys_shutdown",
    294: "sys_setsockopt",
    295: "sys_getsockopt",
    296: "sys_sendmsg",
    297: "sys_recvmsg",
    298: "sys_semop",
    299: "sys_semget",
    300: "sys_semctl",
    301: "sys_msgsnd",
    302: "sys_msgrcv",
    303: "sys_msgget",
    304: "sys_msgctl",
    305: "sys_shmat",
    306: "sys_shmdt",
    307: "sys_shmget",
    308: "sys_shmctl",
    309: "sys_add_key",
    310: "sys_request_key",
    311: "sys_keyctl",
    312: "sys_semtimedop",
    314: "sys_ioprio_set",
    315: "sys_ioprio_get",
    316: "sys_inotify_init",
    317: "sys_inotify_add_watch",
    318: "sys_inotify_rm_watch",
    319: "sys_mbind",
    320: "sys_get_mempolicy",
    321: "sys_set_mempolicy",
    322: "sys_openat",
    323: "sys_mkdirat",
    324: "sys_mknodat",
    325: "sys_fchownat",
    326: "sys_futimesat",
    327: "sys_fstatat64",
    328: "sys_unlinkat",
    329: "sys_renameat",
    330: "sys_linkat",
    331: "sys_symlinkat",
    332: "sys_readlinkat",
    333: "sys_fchmodat",
    334: "sys_faccessat",
    335: "sys_pselect6",
    336: "sys_ppoll",
    337: "sys_unshare",
    338: "sys_set_robust_list",
    339: "sys_get_robust_list",
    340: "sys_splice",
    341: "sys_sync_file_range2",
    342: "sys_tee",
    343: "sys_vmsplice",
    344: "sys_move_pages",
    345: "sys_getcpu",
    346: "sys_epoll_pwait",
    347: "sys_kexec_load",
    348: "sys_utimensat",
    349: "sys_signalfd",
    350: "sys_timerfd_create",
    351: "sys_eventfd",
    352: "sys_fallocate",
    353: "sys_timerfd_settime",
    354: "sys_timerfd_gettime",
    355: "sys_signalfd4",
    356: "sys_eventfd2",
    357: "sys_epoll_create1",
    358: "sys_dup3",
    359: "sys_pipe2",
    360: "sys_inotify_init1",
    361: "sys_preadv",
    362: "sys_pwritev",
    363: "sys_rt_tgsigqueueinfo",
    364: "sys_perf_event_open",
    365: "sys_recvmmsg",
    366: "sys_accept4",
    367: "sys_fanotify_init",
    368: "sys_fanotify_mark",
    369: "sys_prlimit64",
    370: "sys_name_to_handle_at",
    371: "sys_open_by_handle_at",
    372: "sys_clock_adjtime",
    373: "sys_syncfs",
    374: "sys_sendmmsg",
    375: "sys_setns",
    376: "sys_process_vm_readv",
    377: "sys_process_vm_writev",
    378: "sys_kcmp",
    379: "sys_finit_module",
    380: "sys_sched_setattr",
    381: "sys_sched_getattr",
    382: "sys_renameat2",
    383: "sys_seccomp",
    384: "sys_getrandom",
    385: "sys_memfd_create",
    386: "sys_bpf",
    387: "sys_execveat",
    388: "sys_userfaultfd",
    389: "sys_membarrier",
    390: "sys_mlock2",
    391: "sys_copy_file_range",
    392: "sys_preadv2",
    393: "sys_pwritev2",
    394: "sys_pkey_mprotect",
    395: "sys_pkey_alloc",
    396: "sys_pkey_free",
    397: "sys_statx",
    983041: "sys_ARM_NR_breakpoint",
    983042: "sys_ARM_NR_cacheflush",
    983043: "sys_ARM_NR_usr26",
    983044: "sys_ARM_NR_usr32",
    983045: "sys_ARM_NR_set_tls",
}
aarch64 = {
    0: "sys_io_setup",
    1: "sys_io_destroy",
    2: "sys_io_submit",
    3: "sys_io_cancel",
    4: "sys_io_getevents",
    5: "sys_setxattr",
    6: "sys_lsetxattr",
    7: "sys_fsetxattr",
    8: "sys_getxattr",
    9: "sys_lgetxattr",
    10: "sys_fgetxattr",
    11: "sys_listxattr",
    12: "sys_llistxattr",
    13: "sys_flistxattr",
    14: "sys_removexattr",
    15: "sys_lremovexattr",
    16: "sys_fremovexattr",
    17: "sys_getcwd",
    18: "sys_lookup_dcookie",
    19: "sys_eventfd2",
    20: "sys_epoll_create1",
    21: "sys_epoll_ctl",
    22: "sys_epoll_pwait",
    23: "sys_dup",
    24: "sys_dup3",
    25: "sys_fcntl",
    26: "sys_inotify_init1",
    27: "sys_inotify_add_watch",
    28: "sys_inotify_rm_watch",
    29: "sys_ioctl",
    30: "sys_ioprio_set",
    31: "sys_ioprio_get",
    32: "sys_flock",
    33: "sys_mknodat",
    34: "sys_mkdirat",
    35: "sys_unlinkat",
    36: "sys_symlinkat",
    37: "sys_linkat",
    38: "sys_renameat",
    39: "sys_umount",
    40: "sys_mount",
    41: "sys_pivot_root",
    43: "sys_statfs",
    44: "sys_fstatfs",
    45: "sys_truncate",
    46: "sys_ftruncate",
    47: "sys_fallocate",
    48: "sys_faccessat",
    49: "sys_chdir",
    50: "sys_fchdir",
    51: "sys_chroot",
    52: "sys_fchmod",
    53: "sys_fchmodat",
    54: "sys_fchownat",
    55: "sys_fchown",
    56: "sys_openat",
    57: "sys_close",
    58: "sys_vhangup",
    59: "sys_pipe2",
    60: "sys_quotactl",
    61: "sys_getdents64",
    62: "sys_lseek",
    63: "sys_read",
    64: "sys_write",
    65: "sys_readv",
    66: "sys_writev",
    67: "sys_pread64",
    68: "sys_pwrite64",
    69: "sys_preadv",
    70: "sys_pwritev",
    71: "sys_sendfile64",
    72: "sys_pselect6",
    73: "sys_ppoll",
    74: "sys_signalfd4",
    75: "sys_vmsplice",
    76: "sys_splice",
    77: "sys_tee",
    78: "sys_readlinkat",
    79: "sys_newfstatat",
    80: "sys_newfstat",
    81: "sys_sync",
    82: "sys_fsync",
    83: "sys_fdatasync",
    84: "sys_sync_file_range",
    85: "sys_timerfd_create",
    86: "sys_timerfd_settime",
    87: "sys_timerfd_gettime",
    88: "sys_utimensat",
    89: "sys_acct",
    90: "sys_capget",
    91: "sys_capset",
    92: "sys_personality",
    93: "sys_exit",
    94: "sys_exit_group",
    95: "sys_waitid",
    96: "sys_set_tid_address",
    97: "sys_unshare",
    98: "sys_futex",
    99: "sys_set_robust_list",
    100: "sys_get_robust_list",
    101: "sys_nanosleep",
    102: "sys_getitimer",
    103: "sys_setitimer",
    104: "sys_kexec_load",
    105: "sys_init_module",
    106: "sys_delete_module",
    107: "sys_timer_create",
    108: "sys_timer_gettime",
    109: "sys_timer_getoverrun",
    110: "sys_timer_settime",
    111: "sys_timer_delete",
    112: "sys_clock_settime",
    113: "sys_clock_gettime",
    114: "sys_clock_getres",
    115: "sys_clock_nanosleep",
    116: "sys_syslog",
    117: "sys_ptrace",
    118: "sys_sched_setparam",
    119: "sys_sched_setscheduler",
    120: "sys_sched_getscheduler",
    121: "sys_sched_getparam",
    122: "sys_sched_setaffinity",
    123: "sys_sched_getaffinity",
    124: "sys_sched_yield",
    125: "sys_sched_get_priority_max",
    126: "sys_sched_get_priority_min",
    127: "sys_sched_rr_get_interval",
    128: "sys_restart_syscall",
    129: "sys_kill",
    130: "sys_tkill",
    131: "sys_tgkill",
    132: "sys_sigaltstack",
    133: "sys_rt_sigsuspend",
    134: "sys_rt_sigaction",
    135: "sys_rt_sigprocmask",
    136: "sys_rt_sigpending",
    137: "sys_rt_sigtimedwait",
    138: "sys_rt_sigqueueinfo",
    139: "sys_rt_sigreturn",
    140: "sys_setpriority",
    141: "sys_getpriority",
    142: "sys_reboot",
    143: "sys_setregid",
    144: "sys_setgid",
    145: "sys_setreuid",
    146: "sys_setuid",
    147: "sys_setresuid",
    148: "sys_getresuid",
    149: "sys_setresgid",
    150: "sys_getresgid",
    151: "sys_setfsuid",
    152: "sys_setfsgid",
    153: "sys_times",
    154: "sys_setpgid",
    155: "sys_getpgid",
    156: "sys_getsid",
    157: "sys_setsid",
    158: "sys_getgroups",
    159: "sys_setgroups",
    160: "sys_newuname",
    161: "sys_sethostname",
    162: "sys_setdomainname",
    163: "sys_getrlimit",
    164: "sys_setrlimit",
    165: "sys_getrusage",
    166: "sys_umask",
    167: "sys_prctl",
    168: "sys_getcpu",
    169: "sys_gettimeofday",
    170: "sys_settimeofday",
    171: "sys_adjtimex",
    172: "sys_getpid",
    173: "sys_getppid",
    174: "sys_getuid",
    175: "sys_geteuid",
    176: "sys_getgid",
    177: "sys_getegid",
    178: "sys_gettid",
    179: "sys_sysinfo",
    180: "sys_mq_open",
    181: "sys_mq_unlink",
    182: "sys_mq_timedsend",
    183: "sys_mq_timedreceive",
    184: "sys_mq_notify",
    185: "sys_mq_getsetattr",
    186: "sys_msgget",
    187: "sys_msgctl",
    188: "sys_msgrcv",
    189: "sys_msgsnd",
    190: "sys_semget",
    191: "sys_semctl",
    192: "sys_semtimedop",
    193: "sys_semop",
    194: "sys_shmget",
    195: "sys_shmctl",
    196: "sys_shmat",
    197: "sys_shmdt",
    198: "sys_socket",
    199: "sys_socketpair",
    200: "sys_bind",
    201: "sys_listen",
    202: "sys_accept",
    203: "sys_connect",
    204: "sys_getsockname",
    205: "sys_getpeername",
    206: "sys_sendto",
    207: "sys_recvfrom",
    208: "sys_setsockopt",
    209: "sys_getsockopt",
    210: "sys_shutdown",
    211: "sys_sendmsg",
    212: "sys_recvmsg",
    213: "sys_readahead",
    214: "sys_brk",
    215: "sys_munmap",
    216: "sys_mremap",
    217: "sys_add_key",
    218: "sys_request_key",
    219: "sys_keyctl",
    220: "sys_clone",
    221: "sys_execve",
    222: "sys_mmap",
    223: "sys_fadvise64_64",
    224: "sys_swapon",
    225: "sys_swapoff",
    226: "sys_mprotect",
    227: "sys_msync",
    228: "sys_mlock",
    229: "sys_munlock",
    230: "sys_mlockall",
    231: "sys_munlockall",
    232: "sys_mincore",
    233: "sys_madvise",
    234: "sys_remap_file_pages",
    235: "sys_mbind",
    236: "sys_get_mempolicy",
    237: "sys_set_mempolicy",
    238: "sys_migrate_pages",
    239: "sys_move_pages",
    240: "sys_rt_tgsigqueueinfo",
    241: "sys_perf_event_open",
    242: "sys_accept4",
    243: "sys_recvmmsg",
    260: "sys_wait4",
    261: "sys_prlimit64",
    262: "sys_fanotify_init",
    263: "sys_fanotify_mark",
    264: "sys_name_to_handle_at",
    265: "sys_open_by_handle_at",
    266: "sys_clock_adjtime",
    267: "sys_syncfs",
    268: "sys_setns",
    269: "sys_sendmmsg",
    270: "sys_process_vm_readv",
    271: "sys_process_vm_writev",
    272: "sys_kcmp",
    273: "sys_finit_module",
    274: "sys_sched_setattr",
    275: "sys_sched_getattr",
    276: "sys_renameat2",
    277: "sys_seccomp",
    278: "sys_getrandom",
    279: "sys_memfd_create",
    280: "sys_bpf",
    281: "sys_execveat",
    282: "sys_userfaultfd",
    283: "sys_membarrier",
    284: "sys_mlock2",
    285: "sys_copy_file_range",
    286: "sys_preadv2",
    287: "sys_pwritev2",
    288: "sys_pkey_mprotect",
    289: "sys_pkey_alloc",
    290: "sys_pkey_free",
    291: "sys_statx",
}
